package com.htfp.weather.web.service;

import com.aliyun.tablestore.grid.consts.AttributionEnum;
import com.aliyun.tablestore.grid.model.GridDataSet;
import com.aliyun.tablestore.grid.model.GridDataSetMeta;
import com.aliyun.tablestore.grid.model.grid.Grid4D;
import com.htfp.weather.griddata.common.GfsVariableHeightEnum;
import com.htfp.weather.griddata.common.GfsVariableIsobaricEnum;
import com.htfp.weather.griddata.common.TableConfig;
import com.htfp.weather.griddata.operation.GfsDataFetcher;
import com.htfp.weather.griddata.operation.QueryMeta;
import com.htfp.weather.info.Constant;
import com.htfp.weather.info.GfsLevelsEnum;
import com.htfp.weather.utils.DateTimeUtils;
import com.htfp.weather.utils.NdArrayUtils;
import com.htfp.weather.web.exception.AppExcpetion;
import com.htfp.weather.web.exception.ErrorCode;
import com.htfp.weather.web.pojo.response.NowWeatherStatus;
import com.htfp.weather.web.pojo.response.PlaneResponse;
import com.htfp.weather.web.pojo.response.ProfileResponse;
import com.htfp.weather.web.pojo.response.TimeSeriesDataset;
import lombok.extern.slf4j.Slf4j;
import org.springframework.stereotype.Service;
import org.springframework.util.CollectionUtils;
import ucar.ma2.Array;
import ucar.ma2.Index4D;

import javax.annotation.Resource;
import java.lang.reflect.Method;
import java.time.Duration;
import java.time.Instant;
import java.time.OffsetDateTime;
import java.time.ZoneOffset;
import java.time.format.DateTimeFormatter;
import java.util.*;
import java.util.stream.Collectors;

/**
 * @Author : shiyi
 * @Date : 2024/2/2 15:47
 * @Description : 获取tablestore格点数据
 */
@Service("tablestore-gfs")
@Slf4j
public class GfsDataServiceImpl implements IDataService {
    private static final int SURFACE_LEVEL = 9999;
    @Resource
    GfsDataFetcher gfsDataFetcher;
    @Resource
    QueryMeta queryMeta;
    @Resource
    TableConfig tableConfig;


    /**
     * 获取变量随高度的变化廓线
     *
     * @param targetTime   目标时间 UTC
     * @param variableName 数据库中的变量名
     * @param latitude     纬度
     * @param longitude    经度
     * @return {@link ProfileResponse}
     */
    public ProfileResponse getProfile(OffsetDateTime targetTime, String variableName, double latitude, double longitude) {
        String targetVariable = GfsVariableIsobaricEnum.getGfsVariableName(variableName);

        if (targetVariable == null) {
            throw new AppExcpetion(ErrorCode.VALIDATE_ERROR, variableName + "变量在数据库中不存在");
        }

        GridDataSetMeta lastGridDataSetMeta = null;
        try {
            lastGridDataSetMeta = queryMeta.getLastGridDataSetMeta();
        } catch (Exception e) {
            throw new AppExcpetion(ErrorCode.DATA_SET_EMPTY, ": e.getMessage()");
        }
        int iTime = getTargetTimeIndex(targetTime, lastGridDataSetMeta);
        int iLat = getLatitudeIndex(latitude);
        int iLon = getLongitudeIndex(longitude);
        try {
            String gridDataSetId = lastGridDataSetMeta.getGridDataSetId();
            int[] pressureLevels = tableConfig.getPressureList();
            Array array = gfsDataFetcher.getProfile(gridDataSetId, targetVariable, iTime, iLat, iLon);
            float[] values = (float[]) array.getStorage();
            return new ProfileResponse(pressureLevels, values);
        } catch (Exception e) {
            throw new RuntimeException(e);
        }

    }

    /**
     * 获取未来24小时的气象时间序列数据
     *
     * @param latitude  纬度
     * @param longitude 经度
     * @param level     高度：9999代表地面，其他代表指定气压高度
     * @return @{@link TimeSeriesDataset}
     */
    public TimeSeriesDataset getForecastSeries(double latitude, double longitude, int level) {
        GridDataSetMeta lastGridDataSetMeta = null;
        try {
            lastGridDataSetMeta = queryMeta.getLastGridDataSetMeta();
        } catch (Exception e) {
            throw new AppExcpetion(ErrorCode.DATA_SET_EMPTY, ": e.getMessage()");
        }
        Map<String, List> params = getFutureTimeIndexList(lastGridDataSetMeta);
        List<Integer> iTimeList = params.get("iTimeList");
        List<String> timeList = params.get("timeList");
        int[] iTimes = iTimeList.stream().mapToInt(Integer::valueOf).toArray();
        int iLat = getLatitudeIndex(latitude);
        int iLon = getLongitudeIndex(longitude);
        int iLev = getLevelIndex(level);
        List<String> variableNames = getVariableNamesInDatabase(level);

        try {
            String gridDataSetId = lastGridDataSetMeta.getGridDataSetId();
            GridDataSet gridDataSet = gfsDataFetcher.getSeries(gridDataSetId, variableNames, iTimes, iLev, iLat, iLon);

            if (level == SURFACE_LEVEL) {
                return buildTimeSeriesDatasetSurface(timeList, gridDataSet);
            } else {
                return buildTimeSeriesDatasetPressure(timeList, gridDataSet);
            }

        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    /**
     * 获取当前时刻的天气状况
     *
     * @param latitude  纬度
     * @param longitude 经度
     * @param level     高度：9999代表地面，其他代表指定气压高度
     * @return @{@link NowWeatherStatus}
     */
    public NowWeatherStatus getNowWeather(double latitude, double longitude, int level) {
        GridDataSetMeta lastGridDataSetMeta = null;
        try {
            lastGridDataSetMeta = queryMeta.getLastGridDataSetMeta();
        } catch (Exception e) {
            throw new AppExcpetion(ErrorCode.DATA_SET_EMPTY, ": " + e.getMessage());
        }
        int iTime = getTargetTimeIndex(OffsetDateTime.now(), lastGridDataSetMeta);
        int iLat = getLatitudeIndex(latitude);
        int iLon = getLongitudeIndex(longitude);
        int iLev = getLevelIndex(level);
        List<String> variableNames = getVariableNamesInDatabase(level);

        try {
            String gridDataSetId = lastGridDataSetMeta.getGridDataSetId();
            GridDataSet gridDataSet = gfsDataFetcher.getSinglePoint(gridDataSetId, variableNames, iTime, iLev, iLat, iLon);
            // 每个变量都是0维数组
            final GfsLevelsEnum levelFlag = GfsLevelsEnum.getByCode(level);

            NowWeatherStatus nowWeatherStatus = buildNowWeatherInTargetPoint(gridDataSet, 0, 0, 0, levelFlag);
            nowWeatherStatus.setLatitude(latitude);
            nowWeatherStatus.setLongitude(longitude);
            return nowWeatherStatus;
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    /**
     * 获得指定点序列上的当前气象信息
     * @param latitude
     * @param longitude
     * @param level
     * @return
     */
    public List<NowWeatherStatus> getNowWeatherByMultiPoint(List<Double> latitude, List<Double> longitude, int level) {
        GridDataSetMeta lastGridDataSetMeta = null;
        try {
            lastGridDataSetMeta = queryMeta.getLastGridDataSetMeta();
        } catch (Exception e) {
            throw new AppExcpetion(ErrorCode.DATA_SET_EMPTY, ": " + e.getMessage());
        }
        if (latitude.size() != longitude.size()) {
            throw new AppExcpetion(ErrorCode.INDEX_SIZE_ERROR, ": 经纬度数组大小不一致");
        }
        int pointSize = latitude.size();
        List<String> variableNames = getVariableNamesInDatabase(level);
        int iTime = getTargetTimeIndex(OffsetDateTime.now(), lastGridDataSetMeta);
        List<Integer> latIndex = latitude.stream().map(this::getLatitudeIndex).collect(Collectors.toList());
        List<Integer> lonIndex = longitude.stream().map(this::getLongitudeIndex).collect(Collectors.toList());
        int iMinLat = Collections.min(latIndex).intValue();
        int iMaxLat = Collections.max(latIndex).intValue();
        int iMinLon = Collections.min(lonIndex).intValue();
        int iMaxLon = Collections.max(lonIndex).intValue();
        int iLev = getLevelIndex(level);
        int latLength = iMaxLat - iMinLat + 1;
        int lonLength = iMaxLon - iMinLon + 1;
        // int levLength = iMaxLev - iMinLev + 1;
        try {

            String gridDataSetId = lastGridDataSetMeta.getGridDataSetId();
            int[] origin = new int[]{iTime, iLev, 0, 0};
            int[] shape = new int[]{1, 1, tableConfig.getLatSize(), tableConfig.getLatSize()};
            GridDataSet gridDataSet = gfsDataFetcher.queryByTableStore(gridDataSetId, variableNames,
                    origin, shape);

            List<NowWeatherStatus> nowWeatherStatusList = new ArrayList<>();

            for (int i = 0; i < pointSize; i++) {
                NowWeatherStatus nowWeatherStatus = new NowWeatherStatus();
                // NOTE 2024/6/11: build中传入的索引和数据库的查询索引意义不同

                final GfsLevelsEnum levelFlag = GfsLevelsEnum.getByCode(level);
                nowWeatherStatus = buildNowWeatherInTargetPoint(gridDataSet, 0, latIndex.get(i), lonIndex.get(i), levelFlag);
                nowWeatherStatus.setLatitude(latitude.get(i));
                nowWeatherStatus.setLongitude(longitude.get(i));
                nowWeatherStatusList.add(nowWeatherStatus);
            }
            return nowWeatherStatusList;

        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }


    private int[] getMarginIndexByMultiPoint(List<Double> latitude, List<Double> longitude, List<Double> level) {
        int minLat = Integer.MIN_VALUE;
        int maxLat = Integer.MAX_VALUE;
        int minLon = Integer.MIN_VALUE;
        int maxLon = Integer.MAX_VALUE;
        int minLev = Integer.MIN_VALUE;
        int maxLev = Integer.MAX_VALUE;


        int iMinLat = getLatitudeIndex(minLat);
        int iMaxLat = getLatitudeIndex(maxLat);
        int iMinLon = getLongitudeIndex(minLon);
        int iMaxLon = getLongitudeIndex(maxLon);

        return new int[]{iMinLat, iMaxLat, iMinLon, iMaxLon};
    }


    /**
     * 获取指定变量、时间、高度的二维数组数据
     *
     * @param targetTime   目标时间 UTC
     * @param variableName 数据库中的变量名
     * @param level        高度
     * @param minLat       最小纬度
     * @param maxLat       最大纬度
     * @param minLon       最小经度
     * @param maxLon       最大经度
     * @return @{@link PlaneResponse}
     */
    public PlaneResponse getPlane(OffsetDateTime targetTime, String variableName, int level, double minLat, double maxLat, double minLon, double maxLon) {
        int iLev;
        String targetVariable;
        if (level == SURFACE_LEVEL) {
            iLev = 0; // NOTE 2024/5/20: 0 表示地面，风速对应10m高度，其他变量对应2m高度
            targetVariable = GfsVariableHeightEnum.getGfsVariableName(variableName);
        } else {
            iLev = getPressureIndex(level);
            targetVariable = GfsVariableIsobaricEnum.getGfsVariableName(variableName);
        }

        if (targetVariable == null) {
            throw new AppExcpetion(ErrorCode.VALIDATE_ERROR, variableName + "变量在数据库中不存在");
        }
        GridDataSetMeta lastGridDataSetMeta = null;
        try {
            lastGridDataSetMeta = queryMeta.getLastGridDataSetMeta();
        } catch (Exception e) {
            throw new AppExcpetion(ErrorCode.DATA_SET_EMPTY, ": e.getMessage()");
        }

        int iTime = getTargetTimeIndex(targetTime, lastGridDataSetMeta);
        int iMinLat = getLatitudeIndex(minLat);
        int iMaxLat = getLatitudeIndex(maxLat);
        int iMinLon = getLongitudeIndex(minLon);
        int iMaxLon = getLongitudeIndex(maxLon);
        int latLength = iMaxLat - iMinLat + 1;
        int lonLength = iMaxLon - iMinLon + 1;

        try {
            String gridDataSetId = lastGridDataSetMeta.getGridDataSetId();
            Array array = gfsDataFetcher.getPlane(gridDataSetId, targetVariable, iTime, iLev, iMinLat, iMinLon, latLength, lonLength);

            return buildPlane(array, iMinLat, iMinLon, latLength, lonLength);
        } catch (Exception e) {
            throw new RuntimeException(e);
        }

    }


    private List<String> getVariableNamesInDatabase(int level) {
        List<String> variableNames = new ArrayList<>();
        if (level == SURFACE_LEVEL) {
            for (String e : tableConfig.getVariableList()) {
                String gfsVariableName = GfsVariableHeightEnum.getGfsVariableName(e);
                if (gfsVariableName != null) {
                    variableNames.add(gfsVariableName);
                }
            }
        } else {
            for (String e : tableConfig.getVariableList()) {
                String gfsVariableName = GfsVariableIsobaricEnum.getGfsVariableName(e);
                if (gfsVariableName != null) {
                    variableNames.add(gfsVariableName);
                }
            }
        }
        return variableNames;
    }

    private PlaneResponse buildPlane(Array array, int iMinLat, int iMinLon, int latLength, int lonLength) {
        PlaneResponse planeResponse = new PlaneResponse();
        float[][] values = (float[][]) array.reduce().copyToNDJavaArray();
        planeResponse.setLatList(Arrays.copyOfRange(tableConfig.getLatList(), iMinLat, iMinLat + latLength));
        planeResponse.setLonList(Arrays.copyOfRange(tableConfig.getLonList(), iMinLon, iMinLon + lonLength));
        planeResponse.setValues(values);
        return planeResponse;
    }

    private TimeSeriesDataset buildTimeSeriesDatasetPressure(List<String> timeList, GridDataSet gridDataSet) {
        TimeSeriesDataset timeSeriesDataset = new TimeSeriesDataset(timeList.size());
        timeSeriesDataset.setTime(timeList);
        Grid4D grid4D;
        grid4D = gridDataSet.getVariable(GfsVariableIsobaricEnum.getGfsVariableName(GfsVariableIsobaricEnum.TEMP));
        if (grid4D != null) {
            timeSeriesDataset.setTemp((float[]) grid4D.toArray().getStorage());
        }
        grid4D = gridDataSet.getVariable(GfsVariableIsobaricEnum.getGfsVariableName(GfsVariableIsobaricEnum.WIND_SPEED));
        if (grid4D != null) {
            timeSeriesDataset.setWindSpeed((float[]) grid4D.toArray().getStorage());
        }
        grid4D = gridDataSet.getVariable(GfsVariableIsobaricEnum.getGfsVariableName(GfsVariableIsobaricEnum.WIND360));
        if (grid4D != null) {
            timeSeriesDataset.setWind360((float[]) grid4D.toArray().getStorage());
        }
        grid4D = gridDataSet.getVariable(GfsVariableIsobaricEnum.getGfsVariableName(GfsVariableIsobaricEnum.CLOUD));
        if (grid4D != null) {
            timeSeriesDataset.setCloud((float[]) grid4D.toArray().getStorage());
        }
        return timeSeriesDataset;
    }

    private TimeSeriesDataset buildTimeSeriesDatasetSurface(List<String> timeList, GridDataSet gridDataSet) {
        TimeSeriesDataset timeSeriesDataset = new TimeSeriesDataset(timeList.size());
        timeSeriesDataset.setTime(timeList);
        Grid4D grid4D;
        grid4D = gridDataSet.getVariable(GfsVariableHeightEnum.getGfsVariableName(GfsVariableHeightEnum.TEMP));
        if (grid4D != null) {
            timeSeriesDataset.setTemp((float[]) grid4D.toArray().getStorage());
        }
        grid4D = gridDataSet.getVariable(GfsVariableHeightEnum.getGfsVariableName(GfsVariableHeightEnum.WIND_SPEED));
        if (grid4D != null) {
            timeSeriesDataset.setWindSpeed((float[]) grid4D.toArray().getStorage());
        }
        grid4D = gridDataSet.getVariable(GfsVariableHeightEnum.getGfsVariableName(GfsVariableHeightEnum.WIND360));
        if (grid4D != null) {
            timeSeriesDataset.setWind360((float[]) grid4D.toArray().getStorage());
        }
        return timeSeriesDataset;
    }


    /**
     * 构建指定位置的当前天气
     * @param gridDataSet
     * @param iLevInGrid-目标点在返回的网格中的高度索引
     * @param iLatInGrid-目标点在返回的网格中的高度索引
     * @param iLonInGrid-目标点在返回的网格中的高度索引
     * @return
     */
    private NowWeatherStatus buildNowWeatherInTargetPoint(GridDataSet gridDataSet, int iLevInGrid, int iLatInGrid, int iLonInGrid,  GfsLevelsEnum levelFlag) {
        String timeStr = OffsetDateTime.now().withMinute(0).withNano(0).format(DateTimeFormatter.ofPattern(Constant.API_TIME_STRING));
        NowWeatherStatus result = new NowWeatherStatus();
        result.setTime(timeStr);
        // 利用反射构建数据集
        Class<? extends NowWeatherStatus> aClass = result.getClass();
        for (Map.Entry<String, Grid4D> entry : gridDataSet.getVariables().entrySet()) {
            String variableNameInFile = entry.getKey();
            String variableNameInApi;
            if (GfsLevelsEnum.SURFACE.equals(levelFlag)) {
                variableNameInApi = GfsVariableHeightEnum.getVariableNameInApi(variableNameInFile);
            } else {
                variableNameInApi = GfsVariableIsobaricEnum.getVariableNameInApi(variableNameInFile);
            }
            if (variableNameInApi == null) {
                throw new AppExcpetion(ErrorCode.NO_SUCH_VARIABLE);
            }
            Grid4D grid4D = entry.getValue(); Index4D index = new Index4D(grid4D.getShape());
            try {
                String setMethodName = "set" + variableNameInApi.substring(0, 1).toUpperCase() + variableNameInApi.substring(1);
                Method method =  aClass.getDeclaredMethod(setMethodName, aClass.getDeclaredField(variableNameInApi).getType());
                method.invoke(result, grid4D.toArray().getFloat(index.set(0, iLevInGrid, iLatInGrid, iLonInGrid)));
            } catch (ReflectiveOperationException e) {
                throw new AppExcpetion(ErrorCode.NO_SUCH_VARIABLE, e);
            } catch (Exception e) {
                e.printStackTrace();
                throw new AppExcpetion(ErrorCode.RESPONSE_DATA_BUILD_ERROR);
            }
        }
        return result;
    }

    private int getTargetTimeIndex(OffsetDateTime targetTime, GridDataSetMeta lastGridDataSetMeta) {
        long milli = (long) lastGridDataSetMeta.getAttributes().get(AttributionEnum.REFERENCE_TIME.getName());
        OffsetDateTime refTime = OffsetDateTime.ofInstant(Instant.ofEpochMilli(milli), ZoneOffset.ofHours(0));
        if (targetTime.isBefore(refTime)) {
            throw new AppExcpetion(ErrorCode.QUERY_TIME_ERROR);
        }

        List<String> forecastHours = lastGridDataSetMeta.getForecastHours();
        String targetHour = String.valueOf(getForecastHourFromTargetTime(targetTime, refTime));
        if (!forecastHours.contains(targetHour)) {
            throw new AppExcpetion(ErrorCode.QUERY_TIME_ERROR);
        }
        int iTime = forecastHours.indexOf(targetHour);
        return iTime;
    }

    private Map<String, List> getFutureTimeIndexList(GridDataSetMeta lastGridDataSetMeta) {
        long milli = (long) lastGridDataSetMeta.getAttributes().get(AttributionEnum.REFERENCE_TIME.getName());
        OffsetDateTime refTime = OffsetDateTime.ofInstant(Instant.ofEpochMilli(milli), ZoneOffset.ofHours(0));
        OffsetDateTime currentTime = DateTimeUtils.getUTCDateTime(OffsetDateTime.now());
        if (currentTime.isBefore(refTime)) {
            throw new AppExcpetion(ErrorCode.QUERY_TIME_ERROR);
        }
        int startHour = getForecastHourFromTargetTime(currentTime, refTime);
        List<String> forecastHours = lastGridDataSetMeta.getForecastHours();
        List<Integer> iTimeList = new ArrayList<>();
        List<String> timeList = new ArrayList<>();

        for (int i = 0; i < forecastHours.size(); i++) {
            int hour = Integer.parseInt(forecastHours.get(i));
            if (hour >= startHour) {
                iTimeList.add(i);
                String timeStr = DateTimeUtils.getLocalZoneDateTime(refTime.plusHours(hour))
                        .format(DateTimeFormatter.ofPattern(Constant.API_TIME_STRING));
                timeList.add(timeStr);
            }
            if (iTimeList.size() >= 24) {
                break;
            }
        }
        if (CollectionUtils.isEmpty(iTimeList) || CollectionUtils.isEmpty(timeList)) {
            throw new AppExcpetion(ErrorCode.QUERY_TIME_ERROR);
        }
        Map<String, List> map = new HashMap<>(2);
        map.put("iTimeList", iTimeList);
        map.put("timeList", timeList);
        return map;
    }

    private int getLongitudeIndex(double lon) {
        double[] lonList = tableConfig.getLonList();
        if (lonList.length == 0) {
            throw new AppExcpetion(ErrorCode.LONGITUDE_INDEX_ERROR, "经度列表为空， 无法查找索引");
        }
        double minLon = lonList[0];
        double maxLon = lonList[lonList.length - 1];
        if (minLon > lon || lon > maxLon) {
            throw new AppExcpetion(ErrorCode.LONGITUDE_INDEX_ERROR, "经度不在有效范围内: " + minLon + " ~ " + maxLon);
        }
        int nearestIndex = NdArrayUtils.findNearestIndex(lonList, lon);
        if (nearestIndex == -1) {
            throw new AppExcpetion(ErrorCode.LONGITUDE_INDEX_ERROR, "找不到" + lon + "的经度索引");
        }
        return nearestIndex;
    }

    private int getLatitudeIndex(double lat) {
        double[] latList = tableConfig.getLatList();
        if (latList.length == 0) {
            throw new AppExcpetion(ErrorCode.LATITUDE_INDEX_ERROR, "纬度列表为空， 无法查找索引");
        }
        double minLat = latList[0];
        double maxLat = latList[latList.length - 1];

        if (minLat > lat || lat > maxLat) {
            throw new AppExcpetion(ErrorCode.LATITUDE_INDEX_ERROR, "纬度不在有效范围内: " + minLat + " ~ " + maxLat);
        }
        int nearestIndex = NdArrayUtils.findNearestIndex(latList, lat);
        if (nearestIndex == -1) {
            throw new AppExcpetion(ErrorCode.LATITUDE_INDEX_ERROR, "找不到" + lat + "的纬度索引");
        }
        return nearestIndex;
    }

    private int getPressureIndex(int pressureLevel) {
        List<Integer> levList = Arrays.stream(tableConfig.getPressureList()).boxed().collect(Collectors.toList());
        if (CollectionUtils.isEmpty(levList)) {
            throw new AppExcpetion(ErrorCode.LEVEL_INDEX_ERROR, "气压列表为空， 无法查找索引");
        }
        int index = levList.indexOf(pressureLevel);
        if (index == -1) {
            throw new AppExcpetion(ErrorCode.LATITUDE_INDEX_ERROR, "找不到" + pressureLevel + "的气压高度索引");
        }
        return index;
    }

    private int getLevelIndex(int level) {
        if (level == SURFACE_LEVEL) {
            return 0; // NOTE 2024/5/20: 0 表示地面，风速对应10m高度，其他变量对应2m高度
        } else {
            return getPressureIndex(level);
        }
    }

    private int getForecastHourFromTargetTime(OffsetDateTime targetTime, OffsetDateTime refTime) {
        return (int) Duration.between(refTime, targetTime).toHours();
    }


}
